-------------------------------------------------------------------
-- KSI Labs, LLC CENTIPEDE-PCI board for CNC control/whatever.   --
--                                                               --
-- It has 32 optoisolated parallel inputs, 32 open drain outputs --
-- with 5V power for external optocouplers, 10 galvanically      --
-- isolated differential transmitters / 2 receivers (for making  --
-- several SPI interfaces with up to 1 MHz speed or whatever,)   --
-- and 42 unprogrammed I/O lines for extension boards. It is     --
-- supposed to work with CENTIPEDE-BRK breakout board that has   --
-- screw terminals for all signals, 4 signal and 4 power relays, --
-- 5V/3A universal input power supply for feeding secondary side --
-- (i.e. CNC machine side) sensors and whatever else, 2-channel  --
-- SPI DAC (TLV5617A) with 0-10V outputs, and 4-channel SPI ADC  --
-- (TLV1544) with 0-10V inputs.                                  --
--                                                               --
-- More info available from www.ksilabs.com.                     --
-------------------------------------------------------------------

--------------------------------------------------------------------------
--			LICENSE and Terms of Use.			--
--									--
-- First of all, this is NOT GPL or Free Software. It is open so you	--
-- have the VHDL source code. There is also the source code for the	--
-- Windows Engine for Mach3 CNC software available from KSI Labs, LLC	--
-- website, http://www.ksilabs.com.					--
--									--
-- You are free to use this particular VHDL code, make modifications to	--
-- customize it for your own needs, fix inevitable bugs, port it to 	--
-- other CNC software and do anything else as long as your derivative	--
-- work that is based on this code or uses parts or pieces thereof is	--
-- made to be used with any KSI Labs, LLC products.			--
--									--
-- We do NOT grant you a permission to use any parts of this code in	--
-- any products that are designed for products not manufactured or	--
-- distributed by KSI Labs, LLC. We might give such a permission on	--
-- case-by-case basis but it is not legal to use this code or bits and	--
-- pieces thereof with products not coming from KSI Labs, LLC without	--
-- our explicit written permission.					--
--									--
-- The original copyright notices must be included in all derivative	--
-- works. You do NOT have to ask our permission to make any changes to	--
-- this code and you are not under obligation to give us any of your	--
-- work back as long as you are using it for your own needs.		--
--									--
-- However we do NOT give our permission to distribute your derivative	--
-- work unless all original copyrights are retained, all the conditions	--
-- of this license are met (use only with KSI Labs, LLC products) and	--
-- full source code for such derivative work is available for free to	--
-- everybody without any further conditions including but not limited	--
-- to signing any kinds of NDA. We explicitely deny any permission to	--
-- distribute any derivative work based on this code with closed source	--
-- binary-only fragments or include bits or pieces of this code in such	--
-- products. That essentially means that it is NOT legal to use it with --
-- ANY proprietary HDL megafunctions, megacores, etc. if the resulting	--
-- product is distributed. It IS legal to use whatever you want if the	--
-- resulting product is made for use with a KSI Labs, LLC product AND	--
-- that resulting product is not distributed. Exclusions can be granted --
-- by KSI Labs, LLC on case-by-case basis.				--
--------------------------------------------------------------------------

---------------
-- LIBRARIES --
---------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.mach3_data_types.all;
use work.mach3_components.all;

------------
-- ENTITY --
------------

entity centipede_pci is
generic (
	MACH_FIFO_DEPTH		: natural := 20;
	DOUT_INIT_VAL		: std_logic_vector(31 downto 0) := x"00000000";
	DAC_SPI_CLKDIV_VAL	: unsigned(5 downto 0) := b"10_0010";
	ADC_SPI_CLKDIV_VAL	: unsigned(8 downto 0) := b"0_0010_0010";
	ADC_CONVTIME_VAL	: unsigned(8 downto 0) := b"1_1111_0100";
	DOUT_REG_ADR		: std_logic_vector(6 downto 0) := b"000_0000";
	DIN_REG_ADR		: std_logic_vector(6 downto 0) := b"000_0001";
	DAC_REG_ADR		: std_logic_vector(6 downto 0) := b"000_0010";
	ADC_REG_ADR		: std_logic_vector(6 downto 0) := b"000_0011";
	IRQSTAT_REG_ADR		: std_logic_vector(6 downto 0) := b"000_0100";
	IRQMASK_REG_ADR		: std_logic_vector(6 downto 0) := b"000_0101";
	IRQSTAT_MASKED_ADR	: std_logic_vector(6 downto 0) := b"000_0110";
	IDX_0_REG_ADR		: std_logic_vector(6 downto 0) := b"000_0111";
	IDX_1_REG_ADR		: std_logic_vector(6 downto 0) := b"000_1000";
	IDX_2_REG_ADR		: std_logic_vector(6 downto 0) := b"000_1001";
	IDX_3_REG_ADR		: std_logic_vector(6 downto 0) := b"000_1010";
	IDX_4_REG_ADR		: std_logic_vector(6 downto 0) := b"000_1011";
	IDX_5_REG_ADR		: std_logic_vector(6 downto 0) := b"000_1100";
	MACH3CFG_REG_ADR	: std_logic_vector(6 downto 0) := b"000_1101";
	MACH3CFG2_REG_ADR	: std_logic_vector(6 downto 0) := b"000_1110";
	MACH3CTL_REG_ADR	: std_logic_vector(6 downto 0) := b"000_1111";
	MACH3STAT_REG_ADR	: std_logic_vector(6 downto 0) := b"001_0000";
	MACH3FIFO_REG_ADR	: std_logic_vector(6 downto 0) := b"001_0001";
	MACH3ENC_REG_ADR	: std_logic_vector(6 downto 0) := b"001_0010";
	MACH3SPCNT_REG_ADR	: std_logic_vector(6 downto 0) := b"001_0011";
	MACH3STSP_CTL_REG_ADR	: std_logic_vector(6 downto 0) := b"001_0100";
	MACH3CFG3_REG_ADR	: std_logic_vector(6 downto 0) := b"001_0101"
--	EX_0_31_REG_ADR		: std_logic_vector(6 downto 0) := b"001_0101";
--	EX_32_41_REG_ADR	: std_logic_vector(6 downto 0) := b"001_0110";
--	EX_0_31_DIR_REG_ADR	: std_logic_vector(6 downto 0) := b"001_0111";
--	EX_32_41_DIR_REG_ADR	: std_logic_vector(6 downto 0) := b"001_1000"
);
port (
	------------------------
	-- PCI Target 32 bits --
	------------------------

	l_clk		: in std_logic;
	l_rst		: in std_logic;
	l_ad		: inout std_logic_vector(31 downto 0);
	l_cbe		: in std_logic_vector(3 downto 0);
	l_par		: out std_logic;
	l_frame		: in std_logic;
	l_irdy		: in std_logic;
	l_trdy		: out std_logic;
	l_devsel	: out std_logic;
	l_stop		: out std_logic;
	l_idsel		: in std_logic;
	l_perr		: out std_logic;
	l_serr		: out std_logic;
	l_inta		: out std_logic;

	---------
	-- I/O --
	---------

	d_in		: in std_logic_vector(31 downto 0);
	d_out		: out std_logic_vector(31 downto 0);

	---------------
	-- Extension --
	---------------

--	ext		: inout std_logic_vector(41 downto 0);

	-- HINT: Unused SPI pins can be used for making another SPI :)

	----------
	-- SPI1 --
	----------
		
	liso1		: in std_logic;
	losi1		: out std_logic;
	lpck1		: out std_logic;
	lss11		: out std_logic;
	lss12		: out std_logic;
	lss13		: out std_logic;

	----------
	-- SPI2 --
	----------
		
	liso2		: in std_logic;
	losi2		: out std_logic;
	lpck2		: out std_logic;
	lss21		: out std_logic;
	lss22		: out std_logic;
	lss23		: out std_logic;

	----------
	-- Misc --
	----------

	pci_en		: out std_logic
);
end centipede_pci;

------------------
-- ARCHITECTURE --
------------------

architecture guts of centipede_pci is

	----------------
	-- COMPONENTS --
	----------------

	---------------
	-- CONSTANTS --
	---------------
	
	-------------
	-- SIGNALS --
	-------------

	signal clk		: std_logic;
	signal nrst		: std_logic;
	signal rst		: std_logic;
	signal lb_adr		: std_logic_vector(11 downto 0);
	signal lb_din		: std_logic_vector(31 downto 0);
	signal lb_dout		: std_logic_vector(31 downto 0);
	signal lb_nbe		: std_logic_vector(3 downto 0);
	signal lb_wr		: std_logic;
	signal lb_rd		: std_logic;
	signal lb_nint		: std_logic;

--	signal reg_sel		: std_logic_vector(24 downto 0);
	signal reg_sel		: std_logic_vector(21 downto 0);
	signal dout_sel		: std_logic;
	signal din_sel		: std_logic;
	signal dac_sel		: std_logic;
	signal adc_sel		: std_logic;
	signal irq_raw_sel	: std_logic;
	signal irqmask_sel	: std_logic;
	signal irqstat_sel	: std_logic;
	signal mach_idx_sel	: std_logic_vector(5 downto 0);
	signal mach_cfg_sel	: std_logic;
	signal mach_cfg2_sel	: std_logic;
	signal mach_cfg3_sel	: std_logic;
	signal mach_ctl_sel	: std_logic;
	signal mach_stat_sel	: std_logic;
	signal mach_fifo_sel	: std_logic;
	signal mach_enc_sel	: std_logic;
	signal mach_spcnt_sel	: std_logic;
	signal mach_spctl_sel	: std_logic;
	signal wd_restart	: std_logic;
--	signal ex_data_low_sel	: std_logic;
--	signal ex_data_hi_sel	: std_logic;
--	signal ex_dir_low_sel	: std_logic;
--	signal ex_dir_hi_sel	: std_logic;

	signal ioaddr		: std_logic_vector(6 downto 0);

	signal irq_set		: std_logic_vector(14 downto 0);
	signal irq_clr		: std_logic_vector(14 downto 0);
	signal irq		: std_logic;

	signal d_out_reg	: std_logic_vector(31 downto 0);
	signal d_in_reg		: std_logic_vector(31 downto 0);
	signal irqstat_reg	: std_logic_vector(14 downto 0);	
	signal irqmask_reg	: std_logic_vector(14 downto 0);	

--	signal ex_data		: std_logic_vector(41 downto 0);
--	signal ex_dir		: std_logic_vector(41 downto 0);

	signal dac_idle		: std_logic;

	signal adc_rhr_reg	: std_logic_vector(9 downto 0);
	signal adc_idle		: std_logic;
	
	signal pulser_bus	: std_logic_vector(11 downto 0);
	signal axes_active	: std_logic_vector(5 downto 0);
	signal mach_irq_home	: std_logic_vector(5 downto 0);
	signal mach_irq_probe	: std_logic;
	signal mach_irq_cycle	: std_logic;
	signal mach_irq_done	: std_logic;
	signal mach_irq_fault	: std_logic;
	signal mach_irq_fovf	: std_logic;
	signal mach_fovf_clr	: std_logic;
	signal mach_idx		: index_regs;
	signal mach_cfg_reg	: std_logic_vector(31 downto 0);
	signal mach_cfg2_reg	: std_logic_vector(31 downto 0);
	signal mach_cfg3_reg	: std_logic_vector(7 downto 0);
	signal mach_ctl_reg	: std_logic_vector(15 downto 0);
	signal mach_stat_reg	: std_logic_vector(31 downto 0);
	signal mach_enc_reg	: std_logic_vector(31 downto 0);
	signal mach_spcnt_reg	: std_logic_vector(27 downto 0);
	signal mach_spctl_reg	: std_logic_vector(26 downto 0);
	signal mach_enc_data	: std_logic_vector(7 downto 0);

	signal sp_idx_pulse	: std_logic;
	signal sp_tim_pulse	: std_logic;
	signal sp_dirstep	: std_logic;
	signal sp_dir_out	: std_logic;
	signal sp_dir_rev	: std_logic;
	signal sp_step_out	: std_logic;
	signal sp_step_rev	: std_logic;

	signal cpump_train	: std_logic;
	
begin

	--=====================--
	-- COMPONENT INSTANCES --
	--=====================--

	-----------------------------
	-- PCI32TTINY : PCI Target --
	-----------------------------

	u1: component pci32ttiny
	generic map (
		vendorID	=> x"feff",
		deviceID	=> x"0002",
		revisionID	=> x"00",
		subsystemID	=> x"0202",
		subsystemvID	=> x"1172",
		classcodeID	=> x"078000"
	)
	port map (
		-- General 
		pci_clk => l_clk,
		pci_rst => l_rst,
    
		-- PCI target 32bits
		pci_ad => l_ad,
		pci_cbe => l_cbe,
		pci_par => l_par,
		pci_frame => l_frame,
		pci_irdy => l_irdy,
		pci_trdy => l_trdy,
		pci_devsel => l_devsel,
		pci_stop => l_stop,
		pci_idsel => l_idsel,
		pci_perr => l_perr,
		pci_serr => l_serr,
		pci_inta => l_inta,
      
		-- Local bus
		lb_adr => lb_adr,
		lb_data_in => lb_din,
		lb_data_out => lb_dout,
		lb_nbe => lb_nbe,
		lb_nint => lb_nint,
		lb_rd => lb_rd,
		lb_wr => lb_wr
	);

	--------------------------
	-- TLV5617A DAC on SPI1 --
	--------------------------

	u2: component tlv5617a_dac
	generic map (
		clkdiv_val => DAC_SPI_CLKDIV_VAL
	)
	port map (
		dac_idle => dac_idle,
		ss_pin => lss11,
		sclk_pin => lpck1,
		mosi_pin => losi1,
		dac_thr => lb_din(15 downto 0),
		dac_sel => dac_sel,
		wr => lb_wr,
		clk => clk,
		rst => rst
	);

	-------------------------
	-- TLV1544 ADC on SPI2 --
	-------------------------

	u3: component tlv1544_adc
	generic map (
		clkdiv_val => ADC_SPI_CLKDIV_VAL,
		convtime_val => ADC_CONVTIME_VAL
	)
	port map (
		adc_idle => adc_idle,
		adc_rhr => adc_rhr_reg,
		adc_thr => lb_din(9 downto 0),
		ss_pin => lss21,
		sclk_pin => lpck2,
		mosi_pin => losi2,
		miso_pin => liso2,
		adc_sel => adc_sel,
		wr => lb_wr,
		clk => clk,
		rst => rst
	);

	-------------------------------------
	-- Mach 3 Pulsing Engine and stuff --
	-------------------------------------

	u4: component mach3_main
	generic map (
		fifo_depth => MACH_FIFO_DEPTH
	)
	port map (
		pulser_bus => pulser_bus,
		sp_dir => sp_dir_out,
		sp_step => sp_step_out,
		axes_active => axes_active,
		irq_home => mach_irq_home,
		irq_probe => mach_irq_probe,
		irq_cycle => mach_irq_cycle,
		irq_done => mach_irq_done,
		irq_fault => mach_irq_fault,
		irq_fovf => mach_irq_fovf,
		idx => mach_idx,
		cfg_reg => mach_cfg_reg,
		cfg2_reg => mach_cfg2_reg,
		cfg3_reg => mach_cfg3_reg,
		ctl_reg => mach_ctl_reg,
		stat_reg => mach_stat_reg,
		enc_reg => mach_enc_reg,
		spcnt_reg => mach_spcnt_reg,
		stspctl_reg => mach_spctl_reg,
		sp_idx_pulse => sp_idx_pulse,
		sp_tim_pulse => sp_tim_pulse,
		cpump_out => cpump_train,
		fovf_clr => mach_fovf_clr,
		enc_data => mach_enc_data,
		data_in	=> lb_din,
		din_reg	=> d_in_reg,
		idx_sel => mach_idx_sel,
		cfg_sel => mach_cfg_sel,
		cfg2_sel => mach_cfg2_sel,
		cfg3_sel => mach_cfg3_sel,
		ctl_sel => mach_ctl_sel,
		stat_sel => mach_stat_sel,
		fifo_sel => mach_fifo_sel,
		enc_sel => mach_enc_sel,
		spctl_sel => mach_spctl_sel,
		wd_restart => wd_restart,
		rd => lb_rd,
		wr => lb_wr,
		clk => clk,
		nrst => nrst,
		m_rst => rst
	);

	---------------------------
	-- GPIO on Extension Bus --
	---------------------------

--	u5: component extbus_gpio
--	port map (
--		ext_bus => ext,
--		data_in => lb_din,
--		data_out => ex_data,
--		dir_out => ex_dir,
--		data_low_sel => ex_data_low_sel,
--		data_hi_sel => ex_data_hi_sel,
--		dir_low_sel => ex_dir_low_sel,
--		dir_hi_sel => ex_dir_hi_sel,
--		wr => lb_wr,
--		clk => clk,
--		rst => rst
--	);

	----------
	-- MAIN --
	----------

	-----------------------------------------
	-- Latch input data into d_in register --
	-----------------------------------------
	get_d_in : process (clk)
	begin
		if (rising_edge(clk)) then
			d_in_reg <= not d_in;
		end if;
	end process;


	--------------------------------------------------------
	-- Registers read/write from LB/PCI. Only 32 bit size --
	-- access is allowed. All registers are read here but --
	-- write is often done in components where they live. --
	--------------------------------------------------------
	reg_rw : process (nrst, clk)
	begin
		if (nrst='0') then
			d_out_reg <= DOUT_INIT_VAL;
		elsif (rising_edge(clk)) then
			if (dout_sel='1' and lb_rd='1') then
				lb_dout <= d_out_reg;
			elsif (dout_sel='1' and lb_wr='1') then
				d_out_reg <= lb_din;
			elsif (din_sel='1' and lb_rd='1') then
				lb_dout <= d_in_reg;
			elsif (adc_sel='1' and lb_rd='1') then
				lb_dout <= (others => '0');
				lb_dout(adc_rhr_reg'range) <= adc_rhr_reg;
			elsif (irq_raw_sel='1' and lb_rd='1') then
				lb_dout <= (others => '0');
				lb_dout(irqstat_reg'range) <= irqstat_reg;
			elsif (irqmask_sel='1' and lb_rd='1') then
				lb_dout <= (others => '0');
				lb_dout(irqmask_reg'range) <= irqmask_reg;
			elsif (irqmask_sel='1' and lb_wr='1') then
				irqmask_reg <= lb_din(irqmask_reg'range);
			elsif (irqstat_sel='1' and lb_rd='1') then
				lb_dout <= (others => '0');
				lb_dout(irqstat_reg'range) <= irqstat_reg and irqmask_reg;
			elsif (mach_idx_sel(0)='1' and lb_rd='1') then
				lb_dout <= mach_idx(0);
			elsif (mach_idx_sel(1)='1' and lb_rd='1') then
				lb_dout <= mach_idx(1);
			elsif (mach_idx_sel(2)='1' and lb_rd='1') then
				lb_dout <= mach_idx(2);
			elsif (mach_idx_sel(3)='1' and lb_rd='1') then
				lb_dout <= mach_idx(3);
			elsif (mach_idx_sel(4)='1' and lb_rd='1') then
				lb_dout <= mach_idx(4);
			elsif (mach_idx_sel(5)='1' and lb_rd='1') then
				lb_dout <= mach_idx(5);
			elsif (mach_cfg_sel='1' and lb_rd='1') then
				lb_dout <= mach_cfg_reg;
			elsif (mach_cfg2_sel='1' and lb_rd='1') then
				lb_dout <= mach_cfg2_reg;
			elsif (mach_ctl_sel='1' and lb_rd='1') then
				lb_dout <= (others => '0');
				lb_dout(mach_ctl_reg'range) <= mach_ctl_reg;
			elsif (mach_stat_sel='1' and lb_rd='1') then
				lb_dout <= mach_stat_reg;
			elsif (mach_enc_sel='1' and lb_rd='1') then
				lb_dout <= mach_enc_reg;
			elsif (mach_spcnt_sel='1' and lb_rd='1') then
				lb_dout <= (others => '0');
				lb_dout(mach_spcnt_reg'range) <= mach_spcnt_reg;
			elsif (mach_spctl_sel='1' and lb_rd='1') then
				lb_dout <= (others => '0');
				lb_dout(mach_spctl_reg'range) <= mach_spctl_reg;
			elsif (mach_cfg3_sel='1' and lb_rd='1') then
				lb_dout <= (others => '0');
				lb_dout(7 downto 0) <= mach_cfg3_reg;
--			elsif (ex_data_low_sel='1' and lb_rd='1') then
--				lb_dout <= ex_data(31 downto 0);
--			elsif (ex_data_hi_sel='1' and lb_rd='1') then
--				lb_dout <= (others => '0');
--				lb_dout(9 downto 0) <= ex_data(41 downto 32);
--			elsif (ex_dir_low_sel='1' and lb_rd='1') then
--				lb_dout <= ex_dir(31 downto 0);
--			elsif (ex_dir_hi_sel='1' and lb_rd='1') then
--				lb_dout <= (others => '0');
--				lb_dout(9 downto 0) <= ex_dir(41 downto 32);
			end if;
		end if;
	end process;


	----------------------------------------------------------
	-- Top (final) IRQ processing, irq_stat (raw) register. --
	----------------------------------------------------------
	process_irq : process (rst, clk)
	begin
		if (rst='1') then
			irqstat_reg <= (others => '0');
		elsif (rising_edge(clk)) then
			irqstat_reg <= (irqstat_reg or irq_set) and not irq_clr;
		end if;
	end process;


	------------------------------------
	-- Concurrent code, mappings etc. --
	------------------------------------


	-- PCI buffers are always enabled when CPLD is up and running
	pci_en <= '0';

	-- Only one CS per SPI is currently used.
	lss12 <= '0';
	lss13 <= '0';
	lss22 <= '0';
	lss23 <= '0';

	nrst <= l_rst;
	clk <= l_clk;

	mach_enc_data <= d_in_reg(31 downto 24);

	ioaddr <= lb_adr(8 downto 2);

	-- Register select demux, big and ugly :)
	with ioaddr select reg_sel <=
		(0 => '1', others => '0') when DOUT_REG_ADR,
		(1 => '1', others => '0') when DIN_REG_ADR,
		(2 => '1', others => '0') when DAC_REG_ADR,
		(3 => '1', others => '0') when ADC_REG_ADR,
		(4 => '1', others => '0') when IRQSTAT_REG_ADR,
		(5 => '1', others => '0') when IRQMASK_REG_ADR,
		(6 => '1', others => '0') when IRQSTAT_MASKED_ADR,
		(7 => '1', others => '0') when IDX_0_REG_ADR,
		(8 => '1', others => '0') when IDX_1_REG_ADR,
		(9 => '1', others => '0') when IDX_2_REG_ADR,
		(10 => '1', others => '0') when IDX_3_REG_ADR,
		(11 => '1', others => '0') when IDX_4_REG_ADR,
		(12 => '1', others => '0') when IDX_5_REG_ADR,
		(13 => '1', others => '0') when MACH3CFG_REG_ADR,
		(14 => '1', others => '0') when MACH3CFG2_REG_ADR,
		(15 => '1', others => '0') when MACH3CTL_REG_ADR,
		(16 => '1', others => '0') when MACH3STAT_REG_ADR,
		(17 => '1', others => '0') when MACH3FIFO_REG_ADR,
		(18 => '1', others => '0') when MACH3ENC_REG_ADR,
		(19 => '1', others => '0') when MACH3SPCNT_REG_ADR,
		(20 => '1', others => '0') when MACH3STSP_CTL_REG_ADR,
		(21 => '1', others => '0') when MACH3CFG3_REG_ADR,
--		(21 => '1', others => '0') when EX_0_31_REG_ADR,
--		(22 => '1', others => '0') when EX_32_41_REG_ADR,
--		(23 => '1', others => '0') when EX_0_31_DIR_REG_ADR,
--		(24 => '1', others => '0') when EX_32_41_DIR_REG_ADR,
		(others => '0') when others;

	-- Convenience register select signals
	dout_sel <= reg_sel(0);
	din_sel <= reg_sel(1);
	dac_sel <= reg_sel(2);
	adc_sel <= reg_sel(3);
	irq_raw_sel <= reg_sel(4);
	irqmask_sel <= reg_sel(5);
	irqstat_sel <= reg_sel(6);
	mach_idx_sel <= reg_sel(12 downto 7);
	mach_cfg_sel <= reg_sel(13);
	mach_cfg2_sel <= reg_sel(14);
	mach_ctl_sel <= reg_sel(15);
	mach_stat_sel <= reg_sel(16);
	mach_fifo_sel <= reg_sel(17);
	mach_enc_sel <= reg_sel(18);
	mach_spcnt_sel <= reg_sel(19);
	mach_spctl_sel <= reg_sel(20);
	mach_cfg3_sel <= reg_sel(21);
--	ex_data_low_sel <= reg_sel(21);
--	ex_data_hi_sel <= reg_sel(22);
--	ex_dir_low_sel <= reg_sel(23);
--	ex_dir_hi_sel <= reg_sel(24);

	sp_dirstep <= mach_cfg2_reg(8);
	sp_dir_rev <= mach_cfg2_reg(10);
	sp_step_rev <= mach_cfg2_reg(9);

	-- IRQ related
	irq_set <= sp_tim_pulse & sp_idx_pulse & mach_irq_home &
			mach_irq_probe & mach_irq_fovf & mach_irq_fault &
			mach_irq_done & mach_irq_cycle & adc_idle & dac_idle;
	irq_clr <= lb_din(irq_clr'range) when (irq_raw_sel='1' and lb_wr='1') else (others => '0');
	mach_fovf_clr <= irq_clr(5);
	irq <= '1' when ((irqstat_reg and irqmask_reg)=(irqstat_reg'range => '0')) else '0';
	lb_nint <= irq;

	-- Watchdog
	wd_restart <= irq_clr(2);
	
	-- Outputs
	d_out(31 downto 15) <= d_out_reg(31 downto 15);

	axes_map: for i in 0 to 5 generate
		d_out(2*i+1 downto 2*i) <= d_out_reg(2*i+1 downto 2*i) when (axes_active(i)='0') else pulser_bus(2*i+1 downto 2*i);
	end generate;

	d_out(12) <= (sp_dir_out xor sp_dir_rev)  when (sp_dirstep='1') else d_out_reg(12);
	d_out(13) <= (sp_step_out xor sp_step_rev) when (sp_dirstep='1') else d_out_reg(13);

	d_out(14) <= (cpump_train xor mach_cfg3_reg(4)) when (mach_cfg3_reg(1)='1') else d_out_reg(14);

end guts;
