-------------------------------------------------------------------
-- KSI Labs, LLC CENTIPEDE-PCI Mach 3 specific stuff.            --
--                                                               --
-- This is Mach 3 specific stuff. This is main FSM.              --
-------------------------------------------------------------------

--------------------------------------------------------------------------
--			LICENSE and Terms of Use.			--
--									--
-- First of all, this is NOT GPL or Free Software. It is open so you	--
-- have the VHDL source code. There is also the source code for the	--
-- Windows Engine for Mach3 CNC software available from KSI Labs, LLC	--
-- website, http://www.ksilabs.com.					--
--									--
-- You are free to use this particular VHDL code, make modifications to	--
-- customize it for your own needs, fix inevitable bugs, port it to 	--
-- other CNC software and do anything else as long as your derivative	--
-- work that is based on this code or uses parts or pieces thereof is	--
-- made to be used with any KSI Labs, LLC products.			--
--									--
-- We do NOT grant you a permission to use any parts of this code in	--
-- any products that are designed for products not manufactured or	--
-- distributed by KSI Labs, LLC. We might give such a permission on	--
-- case-by-case basis but it is not legal to use this code or bits and	--
-- pieces thereof with products not coming from KSI Labs, LLC without	--
-- our explicit written permission.					--
--									--
-- The original copyright notices must be included in all derivative	--
-- works. You do NOT have to ask our permission to make any changes to	--
-- this code and you are not under obligation to give us any of your	--
-- work back as long as you are using it for your own needs.		--
--									--
-- However we do NOT give our permission to distribute your derivative	--
-- work unless all original copyrights are retained, all the conditions	--
-- of this license are met (use only with KSI Labs, LLC products) and	--
-- full source code for such derivative work is available for free to	--
-- everybody without any further conditions including but not limited	--
-- to signing any kinds of NDA. We explicitely deny any permission to	--
-- distribute any derivative work based on this code with closed source	--
-- binary-only fragments or include bits or pieces of this code in such	--
-- products. That essentially means that it is NOT legal to use it with --
-- ANY proprietary HDL megafunctions, megacores, etc. if the resulting	--
-- product is distributed. It IS legal to use whatever you want if the	--
-- resulting product is made for use with a KSI Labs, LLC product AND	--
-- that resulting product is not distributed. Exclusions can be granted --
-- by KSI Labs, LLC on case-by-case basis.				--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.mach3_data_types.all;


entity mach3_fsm is
port (
	cycle_div	: in std_logic_vector(10 downto 0);
	fifo_pop	: out std_logic;
	fifo_empty	: in std_logic;
	tmr_req		: out std_logic;
	fsm_busy	: out std_logic;
	step_on		: out std_logic;
	step_stb	: out std_logic;
	fsm_run		: in std_logic;
	spindle_pulse	: in std_logic;
	wait4spindle	: in std_logic;
	ctl_wr		: in std_logic;
	clk		: in std_logic;
	rst		: in std_logic
);
end mach3_fsm;

architecture guts of mach3_fsm is

	-------------
	-- Signals --
	-------------

	signal step_start	: std_logic_vector(10 downto 0);

	signal timebase		: unsigned(10 downto 0);

	signal tmr_cntr		: unsigned(2 downto 0);

	signal cy_beg		: std_logic;
	signal cy_end		: std_logic;

	signal step_beg		: std_logic;

	signal busy		: std_logic;

	signal hold4pulse	: std_logic;
	signal hold_prev	: std_logic;
	signal sp_rst		: std_logic;

begin

	hold4pulse_proc : process(rst, fsm_run, clk)
	begin
		if (rst='1' or fsm_run='0') then
			hold4pulse <= '0';
		elsif (rising_edge(clk)) then
			if (wait4spindle='1') then
				hold4pulse <= '1';
			elsif (spindle_pulse='1') then
				hold4pulse <= '0';
			end if;
		end if;
	end process;


	sp_rst_proc : process (clk)
	begin
		if (rising_edge(clk)) then
			hold_prev <= hold4pulse;
		end if;
	end process;


	timebase_proc : process (rst, fsm_run, clk, cycle_div)
	begin
		if (rst='1' or fsm_run='0') then
			timebase <= unsigned(cycle_div);
		elsif (rising_edge(clk)) then
			cy_beg <= '0';
			if ((cy_end or sp_rst)='1') then
				timebase <= unsigned(cycle_div);
				cy_beg <= '1';
			else
				timebase <= timebase - 1;
			end if;
		end if;
	end process;


	step_proc : process (rst, busy, clk)
	begin
		if (rst='1' or busy='0') then
			step_on <= '0';
		elsif (rising_edge(clk)) then
			if ((hold4pulse or cy_end)='1') then
				step_on <= '0';
			elsif (step_beg='1') then
				step_on <= '1';
			end if;
		end if;
	end process;


	tmr_proc : process (rst, fsm_run, clk)
	begin
		if (rst='1' or fsm_run='0') then
			tmr_cntr <= (others => '0');
		elsif (rising_edge(clk)) then
			if (cy_end='1') then
				if (tmr_cntr(2)='1') then
					tmr_cntr <= (others => '0');
				else
					tmr_cntr <= tmr_cntr + 1;
				end if;
			end if;
		end if;
	end process;


	fsm_busy_proc : process (rst, fsm_run, clk)
	begin
		if (rst='1' or fsm_run='0') then
			busy <= '0';
		elsif (rising_edge(clk)) then
			if (fifo_empty='0') then
				busy <= '1';
			elsif (fifo_empty='1' and cy_end='1') then
				busy <= '0';
			end if;
		end if;
	end process;


	step_start <= '0' & cycle_div(10 downto 1);

	cy_end <= '1' when (timebase=(timebase'range => '0')) else '0';

	step_beg <= '1' when (std_logic_vector(timebase)=step_start) else '0';

	step_stb <= step_beg and busy and not hold4pulse;

	fifo_pop <= cy_end and not (fifo_empty or hold4pulse);

	tmr_req <= '1' when (cy_beg='1' and tmr_cntr="000") else '0';

	fsm_busy <= busy;

	sp_rst <= hold_prev and not hold4pulse;

end guts;
